# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for People
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.orderinglist import ordering_list

from .core import Base, uuid_column, Note
from .contact import PhoneNumber, EmailAddress, MailingAddress
from rattail.db.util import normalize_full_name


# TODO: deprecate/remove this
def get_person_display_name(first_name, last_name):
    return normalize_full_name(first_name, last_name)

# TODO: rename this?
def get_person_display_name_from_context(context):
    first_name = context.current_parameters.get('first_name')
    last_name = context.current_parameters.get('last_name')
    return normalize_full_name(first_name, last_name)


class Person(Base):
    """
    Represents a real, living and breathing person.

    (Or, at least was previously living and breathing, in the case of the
    deceased.)
    """
    __tablename__ = 'person'
    __versioned__ = {
        'exclude': ['modified'],
    }

    uuid = uuid_column()
    first_name = sa.Column(sa.String(length=50), nullable=True)
    middle_name = sa.Column(sa.String(length=50), nullable=True)
    last_name = sa.Column(sa.String(length=50), nullable=True)
    # TODO: rename this to 'full_name' at some point...?
    display_name = sa.Column(sa.String(length=100), nullable=True)
    modified = sa.Column(sa.DateTime(), nullable=True, onupdate=datetime.datetime.utcnow)

    users = orm.relationship(
        'User', doc="""
        List of user accounts associated with this person.
        """)

    def __str__(self):
        if six.PY2:
            return unicode(self).encode('utf8')
        return self.display_name or ""

    if six.PY2:
        def __unicode__(self):
            return self.display_name or ""

    def add_email_address(self, address, type='Home'):
        email = PersonEmailAddress(address=address, type=type)
        self.emails.append(email)
        return email

    def add_phone_number(self, number, type='Home'):
        phone = PersonPhoneNumber(number=number, type=type)
        self.phones.append(phone)
        return phone

    def first_valid_email(self):
        """
        Returns the first :class:`Email` which has not been marked invalid, or ``None``.
        """
        for email in self.emails:
            if not email.invalid:
                return email

    def first_valid_email_address(self):
        """
        Returns the first email address which has not been marked invalid, or ``None``.
        """
        email = self.first_valid_email()
        if email:
            return email.address

    def only_customer(self):
        """
        Convenience method to retrieve the one and only ``Customer`` record
        which is associated with this person.  An error will be raised if there
        is not exactly one customer associated.
        """
        if len(self.customers) != 1:
            raise ValueError("person should have 1 customer but instead has {}: {}".format(
                len(self.customers), self))
        return self.customers[0]


class PersonPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a person.
    """

    __mapper_args__ = {'polymorphic_identity': 'Person'}


Person.phones = orm.relationship(
    PersonPhoneNumber,
    primaryjoin=PersonPhoneNumber.parent_uuid == Person.uuid,
    foreign_keys=[PersonPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=PersonPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan',
    doc="""
    Sequence of :class:`PersonPhoneNUmber` instances which belong to the
    person.
    """,
    backref=orm.backref(
        'person',
        doc="""
        Reference to the :class:`Person` instance to which the phone number
        belongs.
        """),
)

Person.phone = orm.relationship(
    PersonPhoneNumber,
    primaryjoin=sa.and_(
        PersonPhoneNumber.parent_uuid == Person.uuid,
        PersonPhoneNumber.preference == 1,
        ),
    foreign_keys=[PersonPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class PersonEmailAddress(EmailAddress):
    """
    Represents an email address associated with a person.
    """

    __mapper_args__ = {'polymorphic_identity': 'Person'}


Person.emails = orm.relationship(
    PersonEmailAddress,
    primaryjoin=PersonEmailAddress.parent_uuid == Person.uuid,
    foreign_keys=[PersonEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=PersonEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan',
    doc="""
    Sequence of :class:`PersonEmailAddress` instances which belong to the
    person.
    """,
    backref=orm.backref(
        'person',
        doc="""
        Reference to the :class:`Person` instance to which the email address
        belongs.
        """),
)

Person.email = orm.relationship(
    PersonEmailAddress,
    primaryjoin=sa.and_(
        PersonEmailAddress.parent_uuid == Person.uuid,
        PersonEmailAddress.preference == 1,
        ),
    foreign_keys=[PersonEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class PersonMailingAddress(MailingAddress):
    """
    Represents a physical / mailing address associated with a person.
    """

    __mapper_args__ = {'polymorphic_identity': 'Person'}


Person.addresses = orm.relationship(
    PersonMailingAddress,
    backref='person',
    primaryjoin=PersonMailingAddress.parent_uuid == Person.uuid,
    foreign_keys=[PersonMailingAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=PersonMailingAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Person.address = orm.relationship(
    PersonMailingAddress,
    primaryjoin=sa.and_(
        PersonMailingAddress.parent_uuid == Person.uuid,
        PersonMailingAddress.preference == 1,
        ),
    foreign_keys=[PersonMailingAddress.parent_uuid],
    uselist=False,
    viewonly=True)


class PersonNote(Note):
    """
    Represents a note attached to a person.
    """
    __mapper_args__ = {'polymorphic_identity': 'Person'}


Person.notes = orm.relationship(
    PersonNote,
    primaryjoin=PersonNote.parent_uuid == Person.uuid,
    foreign_keys=[PersonNote.parent_uuid],
    order_by=PersonNote.created,
    cascade='all, delete-orphan',
    doc="""
    Sequence of notes which belong to the person.
    """,
    backref=orm.backref(
        'person',
        doc="""
        Reference to the person to which the note is attached.
        """))



# -*- coding: utf-8 -*-
"""add handheld batches

Revision ID: 1f8e3cf6b2a2
Revises: 064f71d774ad
Create Date: 2016-08-16 22:16:19.088890

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '1f8e3cf6b2a2'
down_revision = u'064f71d774ad'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # batch_handheld
    op.create_table('batch_handheld',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), batch_id_seq, nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('filename', sa.String(length=255), nullable=False),
                    sa.Column('device_type', sa.String(length=50), nullable=True),
                    sa.Column('device_name', sa.String(length=50), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'batch_handheld_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'batch_handheld_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'batch_handheld_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_handheld_row
    op.create_table('batch_handheld_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('cases', sa.Integer(), nullable=True),
                    sa.Column('units', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'batch_handheld.uuid'], name=u'batch_handheld_row_fk_batch'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'batch_handheld_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_inventory
    op.create_table('batch_inventory',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), batch_id_seq, nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('handheld_batch_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'batch_inventory_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'batch_inventory_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'batch_inventory_fk_executed_by'),
                    sa.ForeignKeyConstraint(['handheld_batch_uuid'], ['batch_handheld.uuid'], name='batch_inventory_fk_handheld_batch'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_inventory_row
    op.create_table('batch_inventory_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('cases', sa.Integer(), nullable=True),
                    sa.Column('units', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'batch_inventory.uuid'], name=u'batch_inventory_row_fk_batch'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'batch_inventory_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # batch_inventory_row
    op.drop_table('batch_inventory_row')

    # batch_inventory
    op.drop_table('batch_inventory')

    # batch_handheld_row
    op.drop_table('batch_handheld_row')

    # batch_handheld
    op.drop_table('batch_handheld')



# -*- coding: utf-8; -*-
# -*- coding: utf-8; -*-
"""
Data model for ${model_title} batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db import model


class ${model_name}(model.FileBatchMixin, model.Base):
    """
    Primary data model for ${model_title} batches.
    """
    __tablename__ = '${table_name}'
    __batchrow_class__ = '${model_name}Row'

    # NOTE: The columns / constraints / relationships shown here are meant to
    # be examples only; please edit as you like.

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='${table_name}_fk_vendor'),
        )

    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)

    vendor = orm.relationship(
        Vendor, doc="""
        Reference to the :class:`Vendor` to which the batch pertains.
        """,
        backref=orm.backref('_${table_name}', cascade='all'))

    effective = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the batch should be considered effective.
    """)


class ${model_name}Row(model.ProductBatchRowMixin, model.Base):
    """
    Row of data within a ${model_title} batch.
    """
    __tablename__ = '${table_name}_row'
    __batch_class__ = ${model_name}

    # NOTE: The columns / constraints / relationships shown here are meant to
    # be examples only; please edit as you like.

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='${table_name}_row_fk_product'),
        )

    STATUS_OK = 1
    STATUS_SOME_CONCERN = 2
    STATUS_UTTER_CHAOS = 3

    STATUS = {
        STATUS_OK:              "groovy",
        STATUS_SOME_CONCERN:    "should look into this one",
        STATUS_UTTER_CHAOS:     "the world just ended",
    }

    upc = sa.Column(model.GPCType(), nullable=True, doc="""
    UPC for the row.
    """)

    product_uuid = sa.Column(sa.String(length=32), nullable=True)

    product = orm.relationship(
        Product,
        doc="""
        Reference to the :class:`Product` with which the row is associated, if any.
        """,
        backref=orm.backref('_${table_name}_rows', cascade='all'))

    description = sa.Column(sa.String(length=255), nullable=False, default='', doc="""
    Description of the product.
    """)

    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Cost per unit of the product.
    """)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SQLAlchemy-Continuum integration
"""

from __future__ import unicode_literals, absolute_import

import socket
import logging

import six

import sqlalchemy as sa
import sqlalchemy_continuum as continuum
from sqlalchemy_continuum.plugins import Plugin
from sqlalchemy_utils.functions import get_primary_keys

from rattail.util import OrderedDict

# TODO: Deprecate/remove this import.
from rattail.db.config import configure_versioning


log = logging.getLogger(__name__)


class RattailPlugin(Plugin):

    def transaction_args(self, uow, session):
        """
        Returns a dict to be used as supplemental kwargs for the
        SQLAlchemy-Continuum transaction.
        """
        user = getattr(session, 'continuum_user', None)

        if hasattr(session, 'continuum_remote_addr'):
            ip = session.continuum_remote_addr
        else:
            host = socket.gethostname()
            ip = socket.gethostbyname(host)

        return {
            'user_id': user.uuid if user else None,
            'remote_addr': ip,
        }


# TODO: we maybe should stop keeping our own reference to this?
# (it used to be "necessary" for customization, not it matters now)
versioning_manager = continuum.versioning_manager


def disable_versioning(manager=None):
    """
    Disable Continuum versioning.
    """
    log.info("disabling Continuum versioning")
    if manager is None:
        manager = versioning_manager
    continuum.remove_versioning(manager=manager)


def count_versions(obj):
    """
    Return the number of versions given object has.

    .. note::
       This function was copied from the one at
       :func:`continuum:sqlalchemy_continuum.utils.count_versions()`, and
       changed so as not to try to use the ``repr()`` value of the table's
       primary key value, since for us that's usually an UUID as unicode string.

    :param obj: SQLAlchemy declarative model object
    """
    session = sa.orm.object_session(obj)
    if session is None:
        # If object is transient, we assume it has no version history.
        return 0
    manager = continuum.get_versioning_manager(obj)
    table_name = manager.option(obj, 'table_name') % obj.__table__.name

    def value(o, k):
        v = getattr(o, k)
        if isinstance(v, six.text_type):
            v = v.encode('utf_8')
        return v

    criteria = [
        '%s = %r' % (pk, value(obj, pk))
        for pk in get_primary_keys(obj)
    ]
    query = 'SELECT COUNT(1) FROM %s WHERE %s' % (
        table_name,
        ' AND '.join(criteria)
    )
    return session.execute(query).scalar()


def model_transaction_query(session, instance, parent_class, child_classes=[]):
    """
    Return a query capable of finding all Continuum ``Transaction`` instances
    which are associated with a model instance.
    """
    from rattail.db import model

    Transaction = continuum.transaction_class(parent_class)
    Version = continuum.version_class(parent_class)

    query = session.query(Transaction)\
        .outerjoin(Version, sa.and_(
            Version.uuid == instance.uuid,
            Version.transaction_id == Transaction.id))

    classes = [Version]
    for model_class, foreign_attr, primary_attr in child_classes:
        cls = continuum.version_class(model_class)
        query = query.outerjoin(cls, sa.and_(
                cls.transaction_id == Transaction.id,
                getattr(cls, foreign_attr) == getattr(instance, primary_attr)))
        classes.append(cls)

    # TODO: i don't recall why we apply this filter, but presumably we'd
    # otherwise get some irrelevant results?  it seems to require at least one
    # of the model records be properly referenced by a version (?)
    query = query.filter(sa.or_(
        *[cls.uuid != None for cls in classes]))

    # we may get multiple results for the same transaction, depending on which
    # child classes were involved etc.  this should set things straight
    query = query.distinct()

    return query



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Products
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.ext.associationproxy import association_proxy

from rattail import enum
from rattail.db.model import (Base, uuid_column, getset_factory,
                              Department, Subdepartment, Category, Vendor, Store)
from rattail.db.types import GPCType
from rattail.db.util import make_full_description
from rattail.util import pretty_quantity


@six.python_2_unicode_compatible
class Brand(Base):
    """
    Represents a brand or similar product line.
    """
    __tablename__ = 'brand'
    __table_args__ = (
        sa.UniqueConstraint('name', name='brand_uq_name'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    name = sa.Column(sa.String(length=100), nullable=False, doc="""
    Brand name, as seen on product packaging etc.
    """)

    confirmed = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the brand has been "confirmed good" by someone responsible
    for product maintenance (if applicable).  In some cases the source of brand
    data is a bit "loose" and it can be helpful to have a human review and
    de-duplicate the brand listing.
    """)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Tax(Base):
    """
    Represents a sales tax rate to be applied to products.
    """
    __tablename__ = 'tax'
    __table_args__ = (
        sa.UniqueConstraint('code', name='tax_uq_code'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    code = sa.Column(sa.String(length=30), nullable=False, doc="""
    Unique "code" for the tax rate.
    """)

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Human-friendly description for the tax rate.
    """)

    rate = sa.Column(sa.Numeric(precision=7, scale=5), nullable=False, doc="""
    Percentage rate for the tax, e.g. 8.25.
    """)

    def __str__(self):
        if self.description:
            return self.description
        return "{} ({}%)".format(self.code, pretty_quantity(self.rate))


class Product(Base):
    """
    Represents a product for sale and/or purchase.
    """
    __tablename__ = 'product'
    __table_args__ = (
        sa.ForeignKeyConstraint(['unit_uuid'], ['product.uuid'], name='product_fk_unit'),
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='product_fk_department'),
        sa.ForeignKeyConstraint(['subdepartment_uuid'], ['subdepartment.uuid'], name='product_fk_subdepartment'),
        sa.ForeignKeyConstraint(['category_uuid'], ['category.uuid'], name='product_fk_category'),
        sa.ForeignKeyConstraint(['family_uuid'], ['family.uuid'], name='product_fk_family'),
        sa.ForeignKeyConstraint(['brand_uuid'], ['brand.uuid'], name='product_fk_brand'),
        sa.ForeignKeyConstraint(['report_code_uuid'], ['report_code.uuid'], name='product_fk_report_code'),
        sa.ForeignKeyConstraint(['deposit_link_uuid'], ['deposit_link.uuid'], name='product_fk_deposit_link'),
        sa.ForeignKeyConstraint(['tax_uuid'], ['tax.uuid'], name='product_fk_tax'),
        sa.ForeignKeyConstraint(['suggested_price_uuid'], ['product_price.uuid'], name='product_fk_suggested_price', use_alter=True),
        sa.ForeignKeyConstraint(['regular_price_uuid'], ['product_price.uuid'], name='product_fk_regular_price', use_alter=True),
        sa.ForeignKeyConstraint(['current_price_uuid'], ['product_price.uuid'], name='product_fk_current_price', use_alter=True),
        sa.Index('product_ix_upc', 'upc'),
    )
    __versioned__ = {'exclude': ['last_sold']}

    uuid = uuid_column()

    upc = sa.Column(GPCType(), nullable=True, doc="""
    Proper GPC value for the product, if any.
    """)

    scancode = sa.Column(sa.String(length=14), nullable=True, doc="""
    Scan code (as string) for the product, if any.  This is generally intended
    to match the :attr:`upc` value, though that is not enforced.
    """)

    item_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Generic ID string for the item.
    """)

    item_type = sa.Column(sa.Integer(), nullable=True, doc="""
    Item type code as integer.
    """)

    department_uuid = sa.Column(sa.String(length=32))
    subdepartment_uuid = sa.Column(sa.String(length=32))
    category_uuid = sa.Column(sa.String(length=32))
    family_uuid = sa.Column(sa.String(length=32))

    report_code_uuid = sa.Column(sa.String(length=32), nullable=True, doc="""
    UUID of the product's report code, if any.
    """)

    deposit_link_uuid = sa.Column(sa.String(length=32), nullable=True, doc="""
    UUID of the product's deposit link, if any.
    """)

    deposit_link = orm.relationship('DepositLink', doc="""
    Reference to the :class:`DepositLink` instance with which the product
    associates, if any.
    """)

    tax_uuid = sa.Column(sa.String(length=32), nullable=True, doc="""
    UUID of the product's tax, if any.
    """)

    tax = orm.relationship(Tax, doc="""
    Reference to the :class:`Tax` instance with which the product associates, if
    any.
    """)

    brand_uuid = sa.Column(sa.String(length=32))

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
    Primary description for the item.
    """)

    description2 = sa.Column(sa.String(length=255), nullable=True, doc="""
    Secondary description for the item.
    """)

    size = sa.Column(sa.String(length=30), nullable=True, doc="""
    Free-form / human-friendly size string for the product, e.g. '32 oz'
    """)

    unit_size = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Unit size for product, as decimal.  This refers to the numeric size of a unit
    of the product, in terms of the :attr:`unit_of_measure`, and may be used for
    smarter price comparisons etc.
    """)

    unit_of_measure = sa.Column(sa.String(length=4), nullable=False,
                                default=enum.UNIT_OF_MEASURE_NONE, doc="""
    Code indicating the unit of measure for the product.  Value should be one of
    the keys of the ``rattail.enum.UNIT_OF_MEASURE`` dictionary.
    """)

    uom_abbreviation = sa.Column(sa.String(length=10), nullable=True, doc="""
    Optional abbreviated unit of measure.  This value is intended to contain a
    human-friendly abbreviation for the :attr:`unit_of_measure`.  This value is
    generally also seen at the end of :attr:`size`.
    """)

    weighed = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether or not the product must be weighed to determine its final price.
    """)

    unit_uuid = sa.Column(sa.String(length=32), nullable=True)

    unit = orm.relationship(
        'Product',
        remote_side=[uuid],
        doc="""
        Reference to the *unit* product upon which this product is based, if any.
        """,
        backref=orm.backref(
            'packs',
            doc="""
            List of products which reference the current product as their unit.
            """))

    pack_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of *units* which constitute the current *pack* product, if applicable.
    """)

    default_pack = sa.Column(sa.Boolean(), nullable=True, doc="""
    If set, this flag indicates the product is the "default pack" for its unit
    item.  This flag is only relevant if the product is in fact a pack item; it
    should be effectively ignored for a unit item.
    """)

    case_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Case size for the product, i.e. how many units (current product) are
    included when *sold* as a case.
    """)

    organic = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether the item is organic.
    """)

    kosher = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether the item is kosher.
    """)

    vegan = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether the item is vegan.
    """)

    vegetarian = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether the item is vegetarian.
    """)

    gluten_free = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether the item is gluten-free.
    """)

    sugar_free = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether the item is sugar-free.
    """)

    not_for_sale = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate items which are not available for sale.
    """)

    discontinued = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate an item has been discontinued.
    """)

    deleted = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag to indicate items which have been deleted.  Obviously this is implies
    "false" deletion, where the record is actually kept on file.  Whether or not
    you use this is up to you.
    """)

    price_required = sa.Column(sa.Boolean(), nullable=True, doc="""
    Indicates that the price must be manually entered by cashier, at check-out.
    """)

    suggested_price_uuid = sa.Column(sa.String(length=32), nullable=True)
    regular_price_uuid = sa.Column(sa.String(length=32), nullable=True)
    current_price_uuid = sa.Column(sa.String(length=32), nullable=True)

    discountable = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
    Whether or not the product may be discounted in any way.
    """)

    special_order = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Whether the product is available for special order.
    """)

    food_stampable = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether food stamps are a valid form of payment for the item.
    """)

    tax1 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 1' applies to the item.
    """)

    tax2 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 2' applies to the item.
    """)

    tax3 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 3' applies to the item.
    """)

    last_sold = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp of the product's last sale event.
    """)

    ingredients = sa.Column(sa.Text(), nullable=True, doc="""
    Free-form ingredients for the product.
    """)

    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
    Status code for the product.  Interpretation of this value is left to the
    custom app logic.
    """)

    notes = sa.Column(sa.Text(), nullable=True, doc="""
    Generic / free-form notes for the product.
    """)

    department = orm.relationship(Department, order_by=Department.name)

    subdepartment = orm.relationship(
        Subdepartment,
        order_by=Subdepartment.name,
        backref=orm.backref('products'))

    category = orm.relationship(
        Category,
        backref=orm.backref('products'))

    brand = orm.relationship(
        Brand,
        doc="""
        Reference to the brand object, if any.
        """,
        backref=orm.backref(
            'products',
            doc="""
            List of products for the brand.
            """))

    family = orm.relationship(
        'Family',
        doc="""
        Reference to the :class:`Family` instance with which the product
        associates, if any.
        """,
        backref=orm.backref(
            'products',
            doc="""
            List of :class:`Product` objects which belong to this family.
            """))

    report_code = orm.relationship(
        'ReportCode',
        doc="""
        Reference to the :class:`ReportCode` instance with which the product
        associates, if any.
        """,
        backref=orm.backref(
            'products',
            doc="""
            List of :class:`Product` objects which associate with this report code.
            """))

    def __str__(self):
        if six.PY2:
            return self.full_description.encode('utf8')
        return self.full_description

    if six.PY2:
        def __unicode__(self):
            return self.full_description

    @property
    def full_description(self):
        """
        Convenience attribute which returns a more complete description.

        Most notably, this includes the brand name and product size.
        """
        return make_full_description(self.brand.name if self.brand else '',
                                     self.description, self.size)

    def is_unit_item(self):
        """
        Returns True if the product is considered a "unit" item, as opposed to
        a "pack" item.
        """
        if self.unit:
            # we appear to be a pack for some other unit item
            return False
        return True

    def is_pack_item(self):
        """
        Returns True if the product is considered a "pack" item, as opposed to
        a "unit" item.
        """
        if self.unit:
            return True
        return False

    @property
    def pretty_upc(self):
        """
        Product's UPC as a somewhat human-readable string.
        """
        if self.upc is None:
            return None
        return self.upc.pretty()

    def costs_for_vendor(self, vendor):
        """
        Return all costs associated with the given vendor.
        """
        return [cost for cost in self.costs
                if cost.vendor is vendor]

    def cost_for_vendor(self, vendor, error_if_multiple=False, error_if_none=False):
        """
        Locate and return the first cost associated with the given vendor.

        :param error_if_multiple: If set, and the product has more than one
           cost associated with the given vendor, an exception will be raised.

        :param error_if_none: If set, and the product has no cost(s) associated
           with the given vendor, an exception will be raised.
        """
        costs = self.costs_for_vendor(vendor)
        if costs:
            if error_if_multiple and len(costs) > 1:
                raise ValueError("Product {} {} has {} costs associated with vendor: {}".format(
                    self.upc, self.full_description, len(costs), vendor))
            return costs[0]
        elif error_if_none:
            raise ValueError("Product {} {} has 0 costs associated with vendor: {}".format(
                self.upc, self.full_description, vendor))


class ProductImage(Base):
    """
    Contains an image for a product.
    """
    __tablename__ = 'product_image'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_image_fk_product'),
    )

    uuid = uuid_column()

    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    product = orm.relationship(
        Product,
        doc="""
        Reference to the product which is shown by the image.
        """,
        backref=orm.backref(
            'image',
            uselist=False,
            doc="""
            Reference to the product's image, if any.
            """))

    bytes = sa.Column(sa.LargeBinary(), nullable=False)


@six.python_2_unicode_compatible
class ProductCode(Base):
    """
    Represents an arbitrary "code" for a product.
    """
    __tablename__ = 'product_code'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_code_fk_product'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)
    code = sa.Column(sa.String(length=20), nullable=True)

    def __str__(self):
        return self.code or ''


Product._codes = orm.relationship(
    ProductCode, backref='product',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=ProductCode.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Product.codes = association_proxy(
    '_codes', 'code',
    getset_factory=getset_factory,
    creator=lambda c: ProductCode(code=c))

Product._code = orm.relationship(
    ProductCode,
    primaryjoin=sa.and_(
        ProductCode.product_uuid == Product.uuid,
        ProductCode.ordinal == 1,
        ),
    uselist=False,
    viewonly=True)

Product.code = association_proxy(
    '_code', 'code',
    getset_factory=getset_factory)


class ProductCost(Base):
    """
    Represents a source from which a product may be obtained via purchase.
    """
    __tablename__ = 'product_cost'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_cost_fk_product'),
        sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='product_cost_fk_vendor'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)

    code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Vendor order code for the item, if applicable.
    """)

    case_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of units which constitute a "case" in the context of this vendor
    catalog item.  May be a fractional quantity, e.g. 17.5LB.
    """)

    # TODO: we're using precision=10, scale=5 elsewhere; should grow this too
    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Case cost for the item, if known.
    """)

    # TODO: we're using precision=10, scale=5 elsewhere; should grow this too
    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Unit cost for the item, if known.
    """)

    pack_size = sa.Column(sa.Integer())
    pack_cost = sa.Column(sa.Numeric(precision=9, scale=5))

    effective = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the cost became effective, if known.
    """)

    discount_starts = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the "current" discount for the cost begins, if
    applicable.  Note that this should never be a "future" value, since only
    the current discount should be part of this record.
    """)

    discount_ends = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the "current" discount for the cost ends, if applicable.
    Note that this should never be a "past" value, since only the current
    discount should be part of this record.
    """)

    discount_amount = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Dollar amount of the "current" discount for the cost, if applicable.
    """)

    discount_percent = sa.Column(sa.Numeric(precision=7, scale=4), nullable=True, doc="""
    Percentage amount of the "current" discount for the cost, if applicable.
    """)

    discount_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Discounted (effective) unit cost for the item, while current discount is in
    effect (if applicable).
    """)

    discontinued = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag to indicate if the cost record has been discontinued, presumably by
    the vendor.
    """)

    vendor = orm.relationship(
        Vendor,
        backref=orm.backref('product_costs', cascade='all'))


Product.costs = orm.relationship(
    ProductCost, backref='product',
    collection_class=ordering_list('preference', count_from=1),
    order_by=ProductCost.preference,
    cascade='save-update, merge, delete, delete-orphan')

Product.cost = orm.relationship(
    ProductCost,
    primaryjoin=sa.and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    uselist=False,
    viewonly=True)

Product.vendor = orm.relationship(
    Vendor,
    secondary=ProductCost.__table__,
    primaryjoin=sa.and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    secondaryjoin=Vendor.uuid == ProductCost.vendor_uuid,
    uselist=False,
    viewonly=True)


class ProductFutureCost(Base):
    """
    Represents a *future* cost record, i.e. an otherwise normal / complete cost
    record, but which will become effective at some later date/time.

    Note that this is meant to be more of a "queue" table in practice,
    i.e. records which exist here should eventually be applied to the
    :class:`ProductCost` table, at which point the future record(s) would be
    deleted.
    """
    __tablename__ = 'product_future_cost'
    __table_args__ = (
        sa.ForeignKeyConstraint(['cost_uuid'], ['product_cost.uuid'], name='product_future_cost_fk_cost'),
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_future_cost_fk_product'),
        sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='product_future_cost_fk_vendor'),
    )

    uuid = uuid_column()

    cost_uuid = sa.Column(sa.String(length=32), nullable=True)
    cost = orm.relationship(
        ProductCost,
        doc="""
        Reference to the "current" cost record to which this future cost record
        will be applied (when the time comes), if applicable.  If this is
        ``None``, then a new "current" cost record would be created instead.
        """,
        backref=orm.backref('futures'))

    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    product = orm.relationship(
        Product,
        doc="""
        Reference to the product to which the cost record pertains.
        """,
        backref=orm.backref(
            'future_costs',
            doc="""
            Sequence of future cost records for the product, i.e. which have
            yet to become "current" costs.
            """))

    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
    vendor = orm.relationship(
        Vendor,
        doc="""
        Reference to the vendor to which the cost record pertains.
        """,
        backref=orm.backref(
            'future_costs',
            doc="""
            Sequence of future cost records for the vendor, i.e. which have yet
            to become "current" costs.
            """))

    # TODO: we don't want this to be part of it, right? (at least not yet)
    # preference = sa.Column(sa.Integer(), nullable=False)

    order_code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Vendor order code for the item, if applicable.
    """)

    case_quantity = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of units which constitute a "case" in the context of this vendor
    catalog item.  May be a fractional quantity, e.g. 17.5LB.
    """)

    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Case cost for the item, if known.
    """)

    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Unit cost for the item, if known.
    """)

    starts = sa.Column(sa.DateTime(), nullable=False, doc="""
    Date and time when the cost becomes effective.
    """)

    ends = sa.Column(sa.DateTime(), nullable=True, doc="""
    Date and time when the cost *stops* being effective, if applicable.  This
    often will be null, in which case the cost becomes "permanently" effective,
    i.e. until a newer cost is brought in.
    """)

    discontinued = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag to indicate if the cost record has been discontinued, presumably by
    the vendor.
    """)


class ProductPrice(Base):
    """
    Represents a price for a product.
    """
    __tablename__ = 'product_price'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_price_fk_product'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    type = sa.Column(sa.Integer())
    level = sa.Column(sa.Integer())
    starts = sa.Column(sa.DateTime())
    ends = sa.Column(sa.DateTime())
    price = sa.Column(sa.Numeric(precision=8, scale=3))
    multiple = sa.Column(sa.Integer())
    pack_price = sa.Column(sa.Numeric(precision=8, scale=3))
    pack_multiple = sa.Column(sa.Integer())

    def active_now(self):
        """
        Returns boolean indicating whether the price is currently active,
        i.e. "now" falls within time window defined by :attr:`starts` and
        :attr:`ends`.  Note that if the price has no start and/or end time,
        this will return ``True``.
        """
        now = datetime.datetime.utcnow()
        if self.starts and self.ends:
            return self.starts <= now <= self.ends
        elif self.starts:
            return self.starts <= now
        elif self.ends:
            return now <= self.ends
        return True


Product.prices = orm.relationship(
    ProductPrice, backref='product',
    primaryjoin=ProductPrice.product_uuid == Product.uuid,
    cascade='save-update, merge, delete, delete-orphan')

Product.suggested_price = orm.relationship(
    ProductPrice,
    primaryjoin=Product.suggested_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.regular_price = orm.relationship(
    ProductPrice,
    primaryjoin=Product.regular_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.current_price = orm.relationship(
    ProductPrice,
    primaryjoin=Product.current_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)


class ProductInventory(Base):
    """
    Inventory data for a product.  Assumption at this point is that this data
    will be accurate only in the context of the "local" node (store etc.).
    Tracking inventory for multiple nodes is not yet supported in the "host"
    sense; however each node may track its own inventory (only).
    """
    __tablename__ = 'product_inventory'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_inventory_fk_product'),
    )

    uuid = uuid_column()

    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    product = orm.relationship(
        Product,
        doc="""
        Product to which this inventory record pertains.
        """,
        backref=orm.backref(
            'inventory',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Inventory data for the product, if any.
            """))

    on_hand = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Unit quantity of product which is currently on hand.
    """)

    on_order = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Unit quantity of product which is currently on order.
    """)


class ProductStoreInfo(Base):
    """
    General store-specific info for a product.
    """
    __tablename__ = 'product_store_info'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_store_info_fk_product'),
        sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'], name='product_store_info_fk_store'),
    )

    uuid = uuid_column()

    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    product = orm.relationship(
        Product,
        doc="""
        Product to which this info record pertains.
        """,
        backref=orm.backref(
            'store_infos',
            collection_class=attribute_mapped_collection('store_uuid'),
            doc="""
            List of store-specific info records for the product.
            """))

    store_uuid = sa.Column(sa.String(length=32), nullable=False)
    store = orm.relationship(
        Store,
        doc="""
        Store to which this info record pertains.
        """)

    recently_active = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating the product has seen "recent activity" at the store.  How
    this is populated and/or interpreted is up to custom app logic.
    """)


class ProductVolatile(Base):
    """
    This is the place to find "volatile" data for a given product, or at least
    it should be...  As of this writing there are a couple other places to look
    but hopefully this table can eventually be "the" place.

    Whether any given value in a given record, applies to the "current" app
    node only, or if it applies to all nodes, is up to app logic.

    Note that a custom app should (most likely) *not* bother "extending" this
    table, but rather should create a separate table with similar pattern.
    """
    __tablename__ = 'product_volatile'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_volatile_fk_product'),
    )

    uuid = uuid_column()

    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    product = orm.relationship(
        Product,
        doc="""
        Product to which this "volatile" data record pertains.
        """,
        backref=orm.backref(
            'volatile',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            "Volatile" data record for the product, if any.
            """))

    true_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    "True" unit cost for the item, if known.  This might include certain
    "allowances" (discounts) currently in effect etc.; really anything which
    might not be reflected in "official" unit cost for the product.  Usually,
    this value is quite easily calculated and so this field serves as more of
    a cache, for sake of SQL access to the values.
    """)

    true_margin = sa.Column(sa.Numeric(precision=8, scale=5), nullable=True, doc="""
    "True" profit margin for the "regular" unit price vs. the "true" unit cost
    (:attr:`true_cost`).
    """)


@six.python_2_unicode_compatible
class InventoryAdjustmentReason(Base):
    """
    Reasons for adjusting product inventory.
    """
    __tablename__ = 'invadjust_reason'
    __table_args__ = (
        sa.UniqueConstraint('code', name='invadjust_reason_uq_code'),
    )

    uuid = uuid_column()

    code = sa.Column(sa.String(length=20), nullable=False, doc="""
    Unique code for the reason.
    """)

    description = sa.Column(sa.String(length=255), nullable=False, doc="""
    Description for the reason.
    """)

    hidden = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the reason code should *not* be generally visible for
    selection by the user etc.
    """)

    def __str__(self):
        return self.description or ""



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from rattail.db import auth, model
from rattail.tests import DataTestCase


class TestAuthenticateUser(DataTestCase):

    def test_nonexistent_user_returns_none(self):
        self.assertTrue(auth.authenticate_user(self.session, u'fred', u'fredpass') is None)

    def test_correct_credentials_returns_user(self):
        fred = model.User(username=u'fred')
        auth.set_user_password(fred, u'fredpass')
        self.session.add(fred)
        self.session.commit()
        user = auth.authenticate_user(self.session, u'fred', u'fredpass')
        self.assertTrue(user is fred)

    def test_wrong_password_user_returns_none(self):
        fred = model.User(username=u'fred', active=False)
        auth.set_user_password(fred, u'fredpass')
        self.session.add(fred)
        self.session.commit()
        self.assertTrue(auth.authenticate_user(self.session, u'fred', u'BADPASS') is None)

    def test_inactive_user_returns_none(self):
        fred = model.User(username=u'fred', active=False)
        auth.set_user_password(fred, u'fredpass')
        self.session.add(fred)
        self.session.commit()
        self.assertTrue(auth.authenticate_user(self.session, u'fred', u'fredpass') is None)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Backup commands
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import subprocess
import logging

import six

from rattail.commands.core import Subcommand


log = logging.getLogger(__name__)


class Backup(Subcommand):
    """
    Backup the database(s) and/or files for this machine
    """
    name = 'backup'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--list-dbs', action='store_true',
                            help="List all databases which may be backed up, then exit.")

        parser.add_argument('--dbdump', action='store_true',
                            help="Dump some or all databases during the backup run.  Exactly "
                            "which databases are dumped, will depend on other parameters.")
        parser.add_argument('--no-dbdump', action='store_true',
                            help="Do NOT dump any databases during the backup run.")

        parser.add_argument('--dbdump-include', action='append', dest='included_dbs', metavar='NAME',
                            help="Name of database to include in the dump.  May be specified "
                            "more than once.  Note that if this parameter is used, then ONLY "
                            "those databases specified will be backed up. Also, the --dbdump-exclude "
                            "parameter(s) will be ignored.")
        parser.add_argument('--dbdump-exclude', action='append', dest='excluded_dbs', metavar='NAME',
                            help="Name of database to exclude from the dump.  May be specified "
                            "more than once.  Note that if the --dbdump-include parameter is specified, "
                            "then this parameter will be ignored.")

        parser.add_argument('--dbdump-output', default='/root/data', metavar='PATH',
                            help="Location of output folder for db dumps.  Default is /root/data")

        parser.add_argument('--rsync', action='store_true',
                            help="Push all files to remote location via rsync.")
        parser.add_argument('--no-rsync', action='store_true',
                            help="Do NOT push files to remote location via rsync.")

        parser.add_argument('--rsync-include', action='append', dest='included_prefixes', metavar='PREFIX',
                            help="File prefix which should be included in the rsync run.  Actually, a "
                            "separate rsync call is made for each prefix.  May be specified more than "
                            "once.  Note that if this parameter is used, then ONLY those prefixes specified "
                            "will be rsync'ed.  Also, the --rsync-exclude parameter(s) will be ignored.")
        parser.add_argument('--rsync-exclude', action='append', dest='excluded_prefixes', metavar='PREFIX',
                            help="File prefix which should be excluded from the rsync run.  May be "
                            "specified more than once.  Note that if the --rsync-include parameter is "
                            "specified, then this parameter will be ignored.")

        parser.add_argument('--rsync-remote-host', metavar='ADDRESS',
                            help="Assuming the rsync destination is another server, this should be the "
                            "hostname or IP address of that server.  Or you could set it to an empty string, "
                            "which signifies that the rsync remote destination is another folder on localhost.")
        parser.add_argument('--rsync-remote-prefix', metavar='PREFIX',
                            help="File prefix for the rsync destination.  If --rsync-remote-host is an empty "
                            "string, this prefix must exist on localhost.  Otherwise the prefix must exist on "
                            "the rsync remote host.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the motions as much as possible, to get an idea "
                            "of what the full backup would do, but don't actually do anything.")

    def run(self, args):

        if args.list_dbs:

            mysql_names = self.get_mysql_names()
            if mysql_names:
                width = max([len(name) for name in mysql_names])
                self.stdout.write("\nmysql\n")
                self.stdout.write("{}\n".format('-' * width))
                for name in mysql_names:
                    self.stdout.write("{}\n".format(name))

            postgres_names = self.get_postgres_names()
            if postgres_names:
                width = max([len(name) for name in postgres_names])
                self.stdout.write("\npostgres\n")
                self.stdout.write("{}\n".format('-' * width))
                for name in postgres_names:
                    self.stdout.write("{}\n".format(name))

            if mysql_names or postgres_names:
                self.stdout.write("\n")
            else:
                self.stdout.write("no databases found\n")

        else: # not listing dbs, so will dump and/or rsync

            if args.dbdump and args.no_dbdump:
                raise RuntimeError("Must specify either --dbdump or --no-dbump, but not both")
            if args.rsync and args.no_rsync:
                raise RuntimeError("Must specify either --rsync or --no-rsync, but not both")

            # we dump dbs by default, unless user or config file says not to
            if not args.no_dbdump and self.config.getbool('rattail.backup', 'dbdump', usedb=False, default=True):
                outdir = args.dbdump_output
                if not os.path.exists(outdir):
                    os.mkdir(outdir)

                if args.included_dbs:
                    include = args.included_dbs
                else:
                    include = self.config.getlist('rattail.backup', 'dbdump.include', usedb=False)

                if args.excluded_dbs:
                    exclude = args.excluded_dbs
                else:
                    exclude = self.config.getlist('rattail.backup', 'dbdump.exclude', usedb=False)

                self.dump_mysql_databases(outdir, include=include, exclude=exclude, dry_run=args.dry_run)
                self.dump_postgres_databases(outdir, include=include, exclude=exclude, dry_run=args.dry_run)

            # we rsync by default, unless user or config file says not to
            if not args.no_rsync and self.config.getbool('rattail.backup', 'rsync', usedb=False, default=True):
                self.rsync_files(include=args.included_prefixes,
                                 exclude=args.excluded_prefixes,
                                 dry_run=args.dry_run)

    def rsync_files(self, include=None, exclude=None, dry_run=False):
        remote_host = self.config.require('rattail.backup', 'rsync.remote_host', usedb=False)
        remote_prefix = self.config.require('rattail.backup', 'rsync.remote_prefix', usedb=False)
        if not remote_prefix.startswith('/'):
            raise RuntimeError("Remote prefix is not absolute path: {}".format(remote_prefix))
        if remote_host:
            remote_prefix = '{}:{}'.format(remote_host, remote_prefix)

        if include:
            prefixes = include
        else:
            prefixes = self.config.getlist('rattail.backup', 'rsync.include', usedb=False, default=[
                '/etc',
                '/home',
                '/opt',
                '/root',
                '/srv',
                '/usr/local',
                '/var',
            ])

        if exclude is None:
            exclude = self.config.getlist('rattail.backup', 'rsync.exclude', usedb=False)

        for prefix in prefixes:
            if not prefix.startswith('/'):
                raise RuntimeError("Prefix is not absolute path: {}".format(prefix))

        for prefix in prefixes:
            others = list(prefixes)
            others.remove(prefix)
            for other in others:
                if other.startswith(prefix):
                    raise RuntimeError("Prefix {} is redundant due to prefix {}".format(
                        other, prefix))

        for prefix in prefixes:
            if not os.path.exists(prefix):
                log.warning("skipping prefix which doesn't exist locally: %s", prefix)
                continue

            excluded = []
            if exclude:
                excluded = [pattern for pattern in exclude if pattern.startswith(prefix)]
            log.info("running rsync for prefix: %s", prefix)

            parent = os.path.dirname(prefix)
            if parent == '/':
                destination = remote_prefix
                excludes = ['--exclude={}'.format(pattern) for pattern in excluded]

            else: # prefix parent is not root (/) dir
                destination = '{}{}/'.format(remote_prefix, parent)

                # exclusion patterns must apparently be relative in this case,
                # i.e. to the topmost folder being synced
                excludes = []
                for pattern in excluded:
                    pattern = '/'.join(pattern.split('/')[2:])
                    excludes.append('--exclude={}'.format(pattern))

                # and must also create the parent folder on remote side
                cmd = ['rsync', '--dirs', '--perms', '--times', '--group', '--owner', parent, remote_prefix]
                log.debug("rsync command is: %s", cmd)
                if not dry_run:
                    subprocess.check_call(cmd)

            # okay let's rsync
            rsync = ['rsync', '--archive', '--delete-during', '--partial']
            if self.verbose:
                rsync.append('--progress')
            cmd = rsync + excludes + [prefix, destination]
            log.debug("rsync command is: %s", cmd)
            if not dry_run:
                subprocess.check_call(cmd)

    def get_mysql_names(self, include=None, exclude=None, exclude_builtin=True):
        try:
            output = subprocess.check_output([
                'mysql', '--execute', "show databases;",
                '--batch', '--skip-column-names',
            ])
        except Exception as error:
            if six.PY3 and isinstance(error, FileNotFoundError):
                # assuming there is no mysql binary
                return []
            raise
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = output.split('\n')
        names = [name.strip() for name in names if name.strip()]

        if exclude_builtin:
            builtins = [
                'mysql',
                'information_schema',
                'performance_schema',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def get_postgres_names(self, include=None, exclude=None, exclude_builtin=True):
        output = subprocess.check_output([
            'sudo', '-u', 'postgres', 'psql', '--list', '--tuples-only',
        ])
        output = output.decode('ascii') # TODO: how to detect this etc.?

        names = []
        for line in output.split('\n'):
            line = line.strip()
            if line and '|' in line and not line.startswith('|'):
                name = line.split('|')[0].strip()
                names.append(name)

        if exclude_builtin:
            builtins = [
                'postgres',
                'template0',
                'template1',
            ]
            for name in builtins:
                if name in names:
                    names.remove(name)

        return self.filter_names(names, include, exclude)

    def filter_names(self, names, include, exclude):
        if include:
            names = [name for name in names if name in include]
        elif exclude:
            names = [name for name in names if name not in exclude]
        return names

    def dump_mysql_databases(self, outdir, include=None, exclude=None, dry_run=False):
        names = self.get_mysql_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping mysql db: %s", name)
            if not dry_run:
                sql_path = os.path.join(outdir, '{}-mysql.sql'.format(name))
                mysqldump = ['mysqldump']
                if self.verbose:
                    mysqldump.append('--verbose')
                cmd = mysqldump + [name, '--result-file={}'.format(sql_path)]
                log.debug("mysqldump command is: %s", cmd)
                subprocess.check_call(cmd)
                subprocess.check_call(['gzip', '--force', sql_path])
                gz_path = "{}.gz".format(sql_path)
                log.debug("dump completed, file is at %s", gz_path)

    def dump_postgres_databases(self, outdir, include=None, exclude=None, dry_run=False):
        names = self.get_postgres_names(include=include, exclude=exclude)
        for name in names:
            log.info("dumping postgres db: %s", name)
            if not dry_run:
                sql_path = os.path.join(outdir, '{}-postgres.sql'.format(name))
                tmp_path = '/tmp/{}.sql'.format(name)
                if os.path.exists(tmp_path):
                    os.remove(tmp_path)
                pg_dump = ['sudo', '-u', 'postgres', 'pg_dump']
                if self.verbose:
                    pg_dump.append('--verbose')
                cmd = pg_dump + [name, '--file={}'.format(tmp_path)]
                log.debug("pg_dump command is: %s", cmd)
                subprocess.check_call(cmd)
                subprocess.check_call(['chown', 'root:', tmp_path])
                subprocess.check_call(['mv', tmp_path, sql_path])
                subprocess.check_call(['gzip', '--force', sql_path])
                gz_path = "{}.gz".format(sql_path)
                log.debug("dump completed, file is at %s", gz_path)



# -*- coding: utf-8 -*-
"""grow inventory.total_cost

Revision ID: eb4e965ce771
Revises: 539c2df56562
Create Date: 2017-09-29 12:01:22.445027

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'eb4e965ce771'
down_revision = u'539c2df56562'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # batch_inventory
    op.alter_column('batch_inventory', 'total_cost', type_=sa.Numeric(precision=12, scale=5))


def downgrade():

    # batch_inventory
    op.alter_column('batch_inventory', 'total_cost', type_=sa.Numeric(precision=9, scale=5))



# -*- coding: utf-8 -*-
"""add category code

Revision ID: 14d83595afa1
Revises: f5c936a4cc62
Create Date: 2016-09-28 15:45:31.331571

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '14d83595afa1'
down_revision = u'f5c936a4cc62'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # category
    op.add_column('category', sa.Column('code', sa.String(length=8), nullable=True))
    op.create_unique_constraint(u'category_uq_code', 'category', ['code'])


def downgrade():

    # category
    op.drop_constraint(u'category_uq_code', 'category', type_='unique')
    op.drop_column('category', 'code')



# -*- coding: utf-8; -*-
"""add product_volatile

Revision ID: 5b393c108673
Revises: 6118198dc4db
Create Date: 2019-03-08 13:14:11.645892

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '5b393c108673'
down_revision = '6118198dc4db'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product_volatile
    op.create_table('product_volatile',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=False),
                    sa.Column('true_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('true_margin', sa.Numeric(precision=8, scale=5), nullable=True),
                    sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_volatile_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # product_volatile
    op.drop_table('product_volatile')



# -*- coding: utf-8; -*-
"""grow pricing_batch_row.markup

Revision ID: fd7cf00b2918
Revises: f3781f071de3
Create Date: 2019-01-08 12:33:51.587887

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'fd7cf00b2918'
down_revision = 'f3781f071de3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'price_markup', type_=sa.Numeric(precision=5, scale=3))


def downgrade():

    # batch_pricing_row
    op.alter_column('batch_pricing_row', 'price_markup', type_=sa.Numeric(precision=4, scale=2))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for pricing batches
"""

from __future__ import unicode_literals, absolute_import

import decimal

import six
from sqlalchemy import orm

from rattail.db import model, api
from rattail.gpc import GPC
from rattail.batch import BatchHandler
from rattail.excel import ExcelReader


class PricingBatchHandler(BatchHandler):
    """
    Handler for pricing batches.
    """
    batch_model_class = model.PricingBatch

    # cached decimal object used for rounding percentages, below
    percent_decimal = decimal.Decimal('.001')

    def should_populate(self, batch):
        if batch.input_filename:
            return True
        if hasattr(batch, 'products'):
            return True
        return False

    def populate(self, batch, progress=None):
        """
        Batch row data comes from product query.
        """
        if batch.input_filename:
            return self.populate_from_file(batch, progress=progress)

        if hasattr(batch, 'product_batch') and batch.product_batch:
            self.populate_from_product_batch(batch, progress=progress)
            return

        assert batch.products
        session = orm.object_session(batch)

        def append(item, i):
            row = model.PricingBatchRow()
            row.product = item
            row.upc = row.product.upc
            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        self.progress_loop(append, batch.products, progress,
                           message="Adding initial rows to batch")

    def populate_from_file(self, batch, progress=None):
        """
        Batch row data comes from input data file.
        """
        path = batch.filepath(self.config, filename=batch.input_filename)
        reader = ExcelReader(path)
        excel_rows = reader.read_rows(progress=progress)
        session = orm.object_session(batch)

        def append(excel, i):
            row = model.PricingBatchRow()

            row.item_entry = excel['upc']
            row.upc = GPC(row.item_entry, calc_check_digit='upc')
            row.product = api.get_product_by_upc(session, row.upc)

            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        self.progress_loop(append, excel_rows, progress,
                           message="Adding initial rows to batch")

    def populate_from_product_batch(self, batch, progress=None):
        """
        Populate pricing batch from product batch.
        """
        session = orm.object_session(batch)
        product_batch = batch.product_batch

        def add(prow, i):
            row = model.PricingBatchRow()
            row.item_entry = prow.item_entry
            with session.no_autoflush:
                row.product = prow.product
            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        self.progress_loop(add, product_batch.active_rows(), progress,
                           message="Adding initial rows to batch")

    def refresh_row(self, row):
        """
        Inspect a row from the source data and populate additional attributes
        for it, according to what we find in the database.
        """
        product = row.product
        if not product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        row.item_id = product.item_id
        row.upc = product.upc
        row.brand_name = six.text_type(product.brand or '')
        row.description = product.description
        row.size = product.size

        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None

        subdept = product.subdepartment
        row.subdepartment_number = subdept.number if subdept else None
        row.subdepartment_name = subdept.name if subdept else None

        family = product.family
        row.family_code = family.code if family else None

        report = product.report_code
        row.report_code = report.code if report else None

        row.alternate_code = product.code

        cost = product.cost
        row.vendor = cost.vendor if cost else None
        row.vendor_item_code = cost.code if cost else None
        row.regular_unit_cost = cost.unit_cost if cost else None

        sugprice = product.suggested_price
        row.suggested_price = sugprice.price if sugprice else None

        regprice = product.regular_price
        row.old_price = regprice.price if regprice else None

    def set_status_per_diff(self, row):
        """
        Set the row's status code according to its price diff
        """
        # prefer "% Diff" if batch defines that
        threshold = row.batch.min_diff_percent
        if threshold:
            # force rounding of row's % diff, for comparison to threshold
            # (this is just to avoid unexpected surprises for the user)
            # (ideally we'd just flush() the session but this seems safer)
            if isinstance(row.price_diff_percent, decimal.Decimal):
                row.price_diff_percent = row.price_diff_percent.quantize(self.percent_decimal)
            # TODO: why don't we use price_diff_percent here again?
            minor = abs(row.margin_diff) < threshold

        else: # or, use "$ Diff" as fallback
            threshold = row.batch.min_diff_threshold
            minor = bool(threshold) and abs(row.price_diff) < threshold

        if row.price_diff > 0:
            if minor:
                row.status_code = row.STATUS_PRICE_INCREASE_MINOR
            else:
                row.status_code = row.STATUS_PRICE_INCREASE
        elif row.price_diff < 0:
            if minor:
                row.status_code = row.STATUS_PRICE_DECREASE_MINOR
            else:
                row.status_code = row.STATUS_PRICE_DECREASE
        else:
            row.status_code = row.STATUS_PRICE_UNCHANGED



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Utilities
"""

from __future__ import unicode_literals, absolute_import
from __future__ import division

import sys
import datetime
import decimal
import subprocess

import six
from pkg_resources import iter_entry_points

try:
    from collections import OrderedDict
except ImportError: # pragma no cover
    from ordereddict import OrderedDict


# generic singleton to indicate an arg which isn't set etc.
NOTSET = object()


def capture_output(command):
    """
    Runs ``command`` and returns any output it produces.
    """
    # We *need* to pipe ``stdout`` because that's how we capture the output of
    # the ``hg`` command.  However, we must pipe *all* handles in order to
    # prevent issues when running as a GUI but *from* the Windows console.  See
    # also: http://bugs.python.org/issue3905
    kwargs = dict(stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = subprocess.Popen(command, **kwargs).communicate()[0]
    return output


def data_diffs(local_data, host_data, fields=None):
    """
    Find all (relevant) fields which differ between the host and local data
    values for a given record.
    """
    if fields is None:
        fields = list(local_data)
    diffs = []
    for field in fields:
        if local_data[field] != host_data[field]:
            diffs.append(field)
    return diffs


def import_module_path(module_path):
    """
    Import an arbitrary Python module.

    :param module_path: String referencing a module by its "dotted path".

    :returns: The referenced module.
    """
    if module_path in sys.modules:
        return sys.modules[module_path]
    module = __import__(module_path)

    def last_module(module, module_path):
        parts = module_path.split('.')
        parts.pop(0)
        child = getattr(module, parts[0])
        if len(parts) == 1:
            return child
        return last_module(child, '.'.join(parts))

    return last_module(module, module_path)


def get_object_spec(obj):
    """
    Returns the "object spec" string for the given object.
    """
    # TODO: this only handles a class *instance* currently
    return '{}:{}'.format(obj.__module__, obj.__class__.__name__)


def load_object(specifier):
    """
    Load an arbitrary object from a module, according to a specifier.

    The specifier string should contain a dotted path to an importable module,
    followed by a colon (``':'``), followed by the name of the object to be
    loaded.  For example:

    .. code-block:: none

       rattail.files:overwriting_move

    You'll notice from this example that "object" in this context refers to any
    valid Python object, i.e. not necessarily a class instance.  The name may
    refer to a class, function, variable etc.  Once the module is imported, the
    ``getattr()`` function is used to obtain a reference to the named object;
    therefore anything supported by that method should work.

    :param specifier: Specifier string.

    :returns: The specified object.
    """
    module_path, name = specifier.split(':')
    module = import_module_path(module_path)
    return getattr(module, name)


def load_entry_points(group):
    """
    Load a set of ``setuptools``-style entry points.

    This is a convenience wrapper around ``pkg_resources.iter_entry_points()``.

    :param group: The group of entry points to be loaded.

    :returns: A dictionary whose keys are the entry point names, and values are
       the loaded entry points.
    """
    entry_points = {}
    for entry_point in iter_entry_points(group):
        entry_points[entry_point.name] = entry_point.load()
    return entry_points


def prettify(text):
    """
    Return a "prettified" version of text.
    """
    text = text.replace('_', ' ')
    text = text.replace('-', ' ')
    words = text.split()
    return ' '.join([x.capitalize() for x in words])


def pretty_boolean(value):
    """
    Returns ``'Yes'`` or ``'No'`` or empty string if value is ``None``
    """
    if value is None:
        return ""
    return "Yes" if value else "No"


def hours_as_decimal(hours, places=2):
    """
    Convert the given ``datetime.timedelta`` object into a Decimal whose
    value is in terms of hours.
    """
    if hours is None:
        return
    minutes = (hours.days * 1440) + (hours.seconds // 60)
    fmt = '{{:0.{}f}}'.format(places)
    return decimal.Decimal(fmt.format(minutes / 60.0))


def pretty_hours(hours=None, seconds=None):
    """
    Format the given ``hours`` value (which is assumed to be a
    ``datetime.timedelta`` object) as HH:MM.  Note that instead of providing
    that, you can provide ``seconds`` instead and a delta will be generated.
    """
    if hours is None and seconds is None:
        return ''
    if hours is None:
        hours = datetime.timedelta(seconds=seconds)

    # determine if hours value is positive or negative.  seems like some of the
    # math can be "off" if negative values are involved, in which case we'll
    # convert to positive for sake of math and then prefix result with a hyphen
    negative = False

    if isinstance(hours, decimal.Decimal):
        if hours < 0:
            negative = True
            hours = -hours
        hours = datetime.timedelta(seconds=int(hours * 3600))

    minutes = (hours.days * 1440) + (hours.seconds / 60)
    rendered = "{}:{:02d}".format(int(minutes // 60), int(minutes % 60))
    if negative:
        rendered = "-{}".format(rendered)
    return rendered


def pretty_quantity(value, empty_zero=False):
    """
    Return a "pretty" version of the given value, as string.  This is meant primarily
    for use with things like order quantities, so that e.g. 1.0000 => 1
    """
    if value is None:
        return ''
    if int(value) == value:
        value = int(value)
        if empty_zero and value == 0:
            return ''
        return six.text_type(value)
    return six.text_type(value).rstrip('0')


def progress_loop(func, items, factory, *args, **kwargs):
    """
    This will iterate over ``items`` and call ``func`` for each.  If a progress
    ``factory`` kwarg is provided, then a progress instance will be created and
    updated along the way.
    """
    message = kwargs.pop('message', None)
    count = kwargs.pop('count', None)
    allow_cancel = kwargs.pop('allow_cancel', False)
    if count is None:
        try:
            count = len(items)
        except TypeError:
            count = items.count()
    if not count:
        return True

    prog = None
    if factory:
        prog = factory(message, count)

    canceled = False
    for i, item in enumerate(items, 1):
        func(item, i, *args, **kwargs)
        if prog and not prog.update(i):
            canceled = True
            break
    if prog:
        prog.finish()
    if canceled and not allow_cancel:
        raise RuntimeError("Operation was canceled")
    return not canceled



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Configuration
"""

from __future__ import unicode_literals, absolute_import

from rattail.config import ConfigProfile, parse_list
from rattail.util import load_object
from rattail.exceptions import ConfigurationError

from rattail.datasync.watchers import NullWatcher


class DataSyncProfile(ConfigProfile):
    """
    Simple class to hold configuration for a DataSync "profile".  Each profile
    determines which database(s) will be watched for new changes, and which
    consumer(s) will then be instructed to process the changes.

    .. todo::
       This clearly needs more documentation.
    """
    section = 'rattail.datasync'

    def __init__(self, config, key):
        self.config = config
        self.key = key

        self.watcher_spec = self._config_string('watcher')
        if self.watcher_spec == 'null':
            self.watcher = NullWatcher(config, key)
        else:
            dbkey = self._config_string('watcher.db', default='default')
            self.watcher = load_object(self.watcher_spec)(config, key, dbkey=dbkey)
        self.watcher.delay = self._config_int('watcher.delay', default=self.watcher.delay)
        self.watcher.retry_attempts = self._config_int('watcher.retry_attempts', default=self.watcher.retry_attempts)
        self.watcher.retry_delay = self._config_int('watcher.retry_delay', default=self.watcher.retry_delay)

        consumers = self._config_list('consumers')
        if consumers == ['self']:
            self.watcher.consumes_self = True
        else:
            self.watcher.consumes_self = False
            self.consumer_delay = self._config_int('consumer_delay', default=1)
            self.consumers = self.normalize_consumers(consumers)
            self.watcher.consumer_stub_keys = [c.key for c in self.consumers]

    def normalize_consumers(self, raw_consumers):
        # TODO: This will do for now, but isn't as awesome as it probably could be.
        default_runas = self._config_string('consumers.runas')
        consumers = []
        for key in raw_consumers:
            consumer_spec = self._config_string('consumer.{}'.format(key))
            if consumer_spec == 'null':
                consumer_spec = 'rattail.datasync.consumers:NullTestConsumer'
            dbkey = self._config_string('consumer.{}.db'.format(key), default=key)
            runas = self._config_string('consumer.{}.runas'.format(key))
            consumer = load_object(consumer_spec)(self.config, key, dbkey=dbkey,
                                                  runas=runas or default_runas,
                                                  watcher=self.watcher)
            consumer.spec = consumer_spec
            consumer.delay = self._config_int(
                'consumer.{}.delay'.format(key),
                default=self.consumer_delay)
            consumer.retry_attempts = self._config_int(
                'consumer.{}.retry_attempts'.format(key),
                default=consumer.retry_attempts)
            consumer.retry_delay = self._config_int(
                'consumer.{}.retry_delay'.format(key),
                default=consumer.retry_delay)
            consumers.append(consumer)
        return consumers


def get_profile_keys(config):
    """
    Returns a list of profile keys used in the DataSync configuration.
    """
    keys = config.get('rattail.datasync', 'watch')
    if keys:
        return parse_list(keys)


def load_profiles(config):
    """
    Load all active DataSync profiles defined within configuration.
    """
    # Make sure we have a top-level directive.
    keys = get_profile_keys(config)
    if not keys:
        raise ConfigurationError(
            "The DataSync configuration does not specify any profiles "
            "to be watched.  Please defined the 'watch' option within "
            "the [rattail.datasync] section of your config file.")

    watched = {}
    for key in keys:
        profile = DataSyncProfile(config, key)
        watched[key] = profile
    return watched



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Definitions
"""

from __future__ import unicode_literals, absolute_import

from openpyxl.styles import PatternFill

from rattail.reporting import Report
from rattail.excel import ExcelWriter


class ExcelReport(Report):
    """
    Generic report which knows how to write Excel output file.

    .. attr:: output_fields

       Simple list of field names which should be included in the output file.

    .. attr:: number_formats

       Optional dictionary specifying "number formats" for any fields.  Use the
       field name for dict key, and value should be the Excel-specific number
       format to be applied to all that column's values.

    .. attr:: totalled_fields

       Optional list of fields which should be "totalled" and represented in a
       final totals row within the output.
    """
    output_fields = []
    number_formats = {}
    totalled_fields = []

    def make_filename(self, session, params, data, **kwargs):
        return "{}.xlsx".format(self.name)

    def write_file(self, session, params, data, path, progress=None, **kwargs):
        """
        Write a basic Excel output file with the given data.  Requires at least
        the ``output_fields`` attribute to be set to work correctly.
        """
        writer = ExcelWriter(path, self.output_fields,
                             number_formats=self.number_formats)
        writer.write_header()

        # convert data to Excel-compatible rows
        rows = [
            [row[field] for field in self.output_fields]
            for row in data]

        # write main data rows
        writer.write_rows(rows)

        # maybe add a TOTALS row
        totals = {}
        for field in self.totalled_fields:
            totals[field] = sum([row[field] for row in data])
        if totals:

            # create totals row data
            rowdata = []
            for field in self.output_fields:
                if field in totals:
                    rowdata.append(totals[field])
                else:
                    rowdata.append(None)

            # append row to output
            writer.write_row(rowdata, row=len(data) + 2)

            # apply row highlighting
            fill_totals = PatternFill(patternType='solid',
                                      fgColor='ffee88',
                                      bgColor='ffee88')
            for col, field in enumerate(self.output_fields, 1):
                cell = writer.sheet.cell(row=len(data) + 2, column=col)
                cell.fill = fill_totals

        writer.auto_freeze()
        writer.auto_filter()
        writer.auto_resize()
        writer.save()



# -*- coding: utf-8; -*-
"""add purchase_batch.order_quantities_known

Revision ID: be079e974a52
Revises: 8b02c6eaf318
Create Date: 2018-07-17 20:16:57.863186

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'be079e974a52'
down_revision = '8b02c6eaf318'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('order_quantities_known', sa.Boolean(), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'order_quantities_known')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail data import
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import Session
from rattail.importing import model
from rattail.importing.handlers import FromSQLAlchemyHandler, ToSQLAlchemyHandler
from rattail.importing.sqlalchemy import FromSQLAlchemy
from rattail.util import OrderedDict


class FromRattailHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    host_title = "Rattail"

    def make_host_session(self):
        return Session()


class ToRattailHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    local_title = "Rattail"

    def make_session(self):
        kwargs = {}
        if hasattr(self, 'runas_user'):
            kwargs['continuum_user'] = self.runas_user
        return Session(**kwargs)

    def begin_local_transaction(self):
        self.session = self.make_session()
        if hasattr(self, 'runas_username') and self.runas_username:
            self.session.set_continuum_user(self.runas_username)


class FromRattailToRattailBase(object):
    """
    Common base class for Rattail -> Rattail data import/export handlers.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['User'] = UserImporter
        importers['AdminUser'] = AdminUserImporter
        importers['Message'] = MessageImporter
        importers['MessageRecipient'] = MessageRecipientImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['Member'] = MemberImporter
        importers['MemberEmailAddress'] = MemberEmailAddressImporter
        importers['MemberPhoneNumber'] = MemberPhoneNumberImporter
        importers['Vendor'] = VendorImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Department'] = DepartmentImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Tax'] = TaxImporter
        importers['InventoryAdjustmentReason'] = InventoryAdjustmentReasonImporter
        importers['Brand'] = BrandImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        importers['ProductStoreInfo'] = ProductStoreInfoImporter
        importers['ProductVolatile'] = ProductVolatileImporter
        importers['ProductImage'] = ProductImageImporter
        importers['LabelProfile'] = LabelProfileImporter
        return importers

    def get_default_keys(self):
        keys = self.get_importer_keys()
        if 'AdminUser' in keys:
            keys.remove('AdminUser')
        if 'ProductImage' in keys:
            keys.remove('ProductImage')
        return keys


class FromRattailToRattailImport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (other) -> Rattail (local) data import.

    .. attribute:: direction

       Value is ``'import'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    dbkey = 'host'

    @property
    def host_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    @property
    def local_title(self):
        return self.config.node_title()

    def make_host_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattailToRattailExport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (local) -> Rattail (other) data export.

    .. attribute:: direction

       Value is ``'export'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    direction = 'export'

    @property
    def host_title(self):
        return self.config.node_title()

    @property
    def local_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    def make_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattail(FromSQLAlchemy):
    """
    Base class for Rattail -> Rattail data importers.
    """

    @property
    def host_model_class(self):
        return self.model_class

    @property
    def supported_fields(self):
        """
        We only need to support the simple fields in a Rattail->Rattail import,
        since all relevant tables should be covered and therefore no need to do
        crazy foreign key acrobatics etc.
        """
        return self.simple_fields

    def query(self):
        """
        Leverage the same caching optimizations on both sides, if applicable.
        """
        query = super(FromRattail, self).query()
        if hasattr(self, 'cache_query_options'):
            options = self.cache_query_options()
            if options:
                for option in options:
                    query = query.options(option)
        return query

    def normalize_host_object(self, obj):
        """
        Normalization should work the same for both sides.
        """
        return self.normalize_local_object(obj)


class PersonImporter(FromRattail, model.PersonImporter):
    pass

class PersonEmailAddressImporter(FromRattail, model.PersonEmailAddressImporter):
    pass

class PersonPhoneNumberImporter(FromRattail, model.PersonPhoneNumberImporter):
    pass

class PersonMailingAddressImporter(FromRattail, model.PersonMailingAddressImporter):
    pass

class UserImporter(FromRattail, model.UserImporter):
    pass

class AdminUserImporter(FromRattail, model.AdminUserImporter):

    @property
    def supported_fields(self):
        return super(AdminUserImporter, self).supported_fields + [
            'admin',
        ]

    def normalize_host_object(self, user):
        data = super(AdminUserImporter, self).normalize_local_object(user) # sic
        if 'admin' in self.fields: # TODO: do we really need this, after the above?
            data['admin'] = self.admin_uuid in [r.role_uuid for r in user._roles]
        return data


class MessageImporter(FromRattail, model.MessageImporter):
    pass

class MessageRecipientImporter(FromRattail, model.MessageRecipientImporter):
    pass

class StoreImporter(FromRattail, model.StoreImporter):
    pass

class StorePhoneNumberImporter(FromRattail, model.StorePhoneNumberImporter):
    pass

class EmployeeImporter(FromRattail, model.EmployeeImporter):
    pass

class EmployeeStoreImporter(FromRattail, model.EmployeeStoreImporter):
    pass

class EmployeeDepartmentImporter(FromRattail, model.EmployeeDepartmentImporter):
    pass

class EmployeeEmailAddressImporter(FromRattail, model.EmployeeEmailAddressImporter):
    pass

class EmployeePhoneNumberImporter(FromRattail, model.EmployeePhoneNumberImporter):
    pass

class ScheduledShiftImporter(FromRattail, model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(FromRattail, model.WorkedShiftImporter):
    pass

class CustomerImporter(FromRattail, model.CustomerImporter):
    pass

class CustomerGroupImporter(FromRattail, model.CustomerGroupImporter):
    pass

class CustomerGroupAssignmentImporter(FromRattail, model.CustomerGroupAssignmentImporter):
    pass

class CustomerPersonImporter(FromRattail, model.CustomerPersonImporter):
    pass

class CustomerEmailAddressImporter(FromRattail, model.CustomerEmailAddressImporter):
    pass

class CustomerPhoneNumberImporter(FromRattail, model.CustomerPhoneNumberImporter):
    pass

class MemberImporter(FromRattail, model.MemberImporter):
    pass

class MemberEmailAddressImporter(FromRattail, model.MemberEmailAddressImporter):
    pass

class MemberPhoneNumberImporter(FromRattail, model.MemberPhoneNumberImporter):
    pass

class VendorImporter(FromRattail, model.VendorImporter):
    pass

class VendorEmailAddressImporter(FromRattail, model.VendorEmailAddressImporter):
    pass

class VendorPhoneNumberImporter(FromRattail, model.VendorPhoneNumberImporter):
    pass

class VendorContactImporter(FromRattail, model.VendorContactImporter):
    pass

class DepartmentImporter(FromRattail, model.DepartmentImporter):
    pass

class SubdepartmentImporter(FromRattail, model.SubdepartmentImporter):
    pass

class CategoryImporter(FromRattail, model.CategoryImporter):
    pass

class FamilyImporter(FromRattail, model.FamilyImporter):
    pass

class ReportCodeImporter(FromRattail, model.ReportCodeImporter):
    pass

class DepositLinkImporter(FromRattail, model.DepositLinkImporter):
    pass

class TaxImporter(FromRattail, model.TaxImporter):
    pass

class InventoryAdjustmentReasonImporter(FromRattail, model.InventoryAdjustmentReasonImporter):
    pass

class BrandImporter(FromRattail, model.BrandImporter):
    pass

class ProductImporter(FromRattail, model.ProductImporter):

    # TODO...
    @property
    def simple_fields(self):
        fields = super(ProductImporter, self).simple_fields
        fields.remove('regular_price_uuid')
        fields.remove('current_price_uuid')
        return fields

    def query(self):
        query = super(ProductImporter, self).query()

        # make sure potential unit items (i.e. rows with NULL unit_uuid) come
        # first, so they will be created before pack items reference them
        # cf. https://www.postgresql.org/docs/current/static/queries-order.html
        # cf. https://stackoverflow.com/a/7622046
        query = query.order_by(self.host_model_class.unit_uuid.desc())

        return query


class ProductCodeImporter(FromRattail, model.ProductCodeImporter):
    pass

class ProductCostImporter(FromRattail, model.ProductCostImporter):
    pass

class ProductPriceImporter(FromRattail, model.ProductPriceImporter):

    @property
    def supported_fields(self):
        return super(ProductPriceImporter, self).supported_fields + self.product_reference_fields


class ProductStoreInfoImporter(FromRattail, model.ProductStoreInfoImporter):
    pass

class ProductVolatileImporter(FromRattail, model.ProductVolatileImporter):
    pass


class ProductImageImporter(FromRattail, model.ProductImageImporter):
    """
    Importer for product images.  Note that this uses the "batch" approach
    because fetching all data up front is not performant when the host/local
    systems are on different machines etc.
    """

    def query(self):
        query = self.host_session.query(self.model_class)\
                                 .order_by(self.model_class.uuid)
        return query[self.host_index:self.host_index + self.batch_size]


class LabelProfileImporter(FromRattail, model.LabelProfileImporter):

    def query(self):
        query = super(LabelProfileImporter, self).query()

        if not self.config.getbool('rattail', 'labels.sync_all_profiles', default=False):
            # only fetch labels from host which are marked as "sync me"
            query = query .filter(self.model_class.sync_me == True)

        return query.order_by(self.model_class.ordinal)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for generic data export history
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import uuid_column, User


@six.python_2_unicode_compatible
class ExportMixin(object):
    """
    Mixin for all generic export models
    """

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__()

    @classmethod
    def __default_table_args__(cls):
        return cls.__export_table_args__()

    @classmethod
    def __export_table_args__(cls):
        return (
            sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'],
                                    name='{}_fk_created_by'.format(cls.__tablename__)),
        )

    @declared_attr
    def export_key(cls):
        return cls.__tablename__

    uuid = uuid_column()

    id = sa.Column(sa.Integer(), sa.Sequence('batch_id_seq'), nullable=False, doc="""
    Unique ID number for the export.
    """)

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date and time when the product export was created.
    """)

    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def created_by(cls):
        return orm.relationship(
            User,
            doc="""
            Reference to the user who created the export.
            """)
    
    record_count = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of records included in the export.  This is a cached value,
    hopefully set at the time the export is generated.
    """)

    def __str__(self):
        return self.id_str

    @property
    def id_str(self):
        if self.id:
            return '{:08d}'.format(self.id)
        return 'pending'

    def filepath(self, config, filename=None, makedirs=False):
        """
        Returns absolute path to export's data folder, optionally appended with
        given filename.  This is a convenience method which really just invokes
        :meth:`rattail.config.RattailConfig.export_filepath()`.
        """
        return config.export_filepath(self.export_key, self.uuid,
                                      filename=filename, makedirs=makedirs)



# -*- coding: utf-8 -*-
"""make tax code a string

Revision ID: c2af678f010c
Revises: 6a890bd49ad0
Create Date: 2017-02-06 17:40:17.486771

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'c2af678f010c'
down_revision = u'a48db300576a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # tax
    op.alter_column('tax', 'code', type_=sa.String(length=30))


def downgrade():

    # tax
    op.alter_column('tax', 'code', type_=sa.Integer())



# -*- coding: utf-8 -*-
"""add transaction_meta

Revision ID: 65098d24972f
Revises: 3b2859694c53
Create Date: 2017-07-04 15:19:50.832960

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '65098d24972f'
down_revision = u'3b2859694c53'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # transaction_meta
    op.create_table('transaction_meta',
                    sa.Column('transaction_id', sa.BigInteger(), nullable=False),
                    sa.Column('key', sa.Unicode(length=255), nullable=False),
                    sa.Column('value', sa.UnicodeText(), nullable=True),
                    sa.PrimaryKeyConstraint('transaction_id', 'key')
    )


def downgrade():

    # transaction_meta
    op.drop_table('transaction_meta')
